
import lightgbm as lgb
import numpy as np
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
import tensorflow as tf
import tensorflow.keras as k
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras.callbacks import ModelCheckpoint
from sklearn.metrics import roc_curve


lgb_params = {
    'task': 'train',
    'boosting_type': 'gbdt',
    'objective': 'binary',
    'metric': ['binary_logloss', 'auc'],
    'learning_rate': 0.005,
    'feature_fraction': 0.9,
    'bagging_fraction': 0.7,
    'bagging_freq': 10,
    'verbose': -1,
    "max_depth": 8,
    "num_leaves": 128,  
    "max_bin": 512,
    "force_row_wise": True,
    'nthread': 30,
    #'is_unbalance': True
}

def build_classification_model_1(X_train, y_train, X_test, y_test, epochs=100, lgb_params=lgb_params):
    

    d_train=lgb.Dataset(X_train, label=y_train, params={'verbose': -1})
    d_test=lgb.Dataset(X_test, label=y_test, params={'verbose': -1})
    
         
    evals_results = {}
    
    early_stopping_rounds = 15000
    
    
    clf = lgb.train(lgb_params, 
                     d_train, 
                     valid_sets=[d_train, d_test], 
                     valid_names=['train', 'test'], 
                     evals_result=evals_results, 
                     num_boost_round=epochs,
                     early_stopping_rounds=early_stopping_rounds,
                     verbose_eval=-1,
                     feval=None)
    
    y_pred = clf.predict(X_train)
    
    fpr, tpr, thresholds = roc_curve(y_train, y_pred)
    # calculate the g-mean for each threshold
    gmeans = np.sqrt(tpr * (1-fpr))
    # locate the index of the largest g-mean
    ix = np.argmax(gmeans)
    print('Best Threshold=%f, G-Mean=%.3f' % (thresholds[ix], gmeans[ix]))
        
        
    return clf, thresholds[ix]



import tensorflow.keras as k

def build_classification_model_2(X_train_s, y_train, X_test_s, y_test, epochs=100):
    

    METRICS = [
          #k.metrics.Accuracy(name='Accuracy'),
          k.metrics.AUC(name='AUC'),
          k.metrics.Precision(name='Precision')
    ]

    def make_model(metrics=METRICS, output_bias=None):
        if output_bias is not None:
            output_bias = tf.keras.initializers.Constant(output_bias)
        model = k.Sequential([
            k.layers.Dense(
                  64, activation='relu',
                  input_shape=(X_train_s.shape[-1],)),
            k.layers.Dropout(0.2),

            k.layers.Dense(
                  32, activation='relu'),
            k.layers.Dropout(0.1),

            k.layers.Dense(
                  16, activation='relu'),

            k.layers.Dense(1, activation='sigmoid',
                                 bias_initializer=output_bias),
            ])

        model.compile(
              optimizer=k.optimizers.Adam(lr=1e-3),
              loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
              metrics=metrics)

        return model
    

    BATCH_SIZE = 32

    early_stopping = tf.keras.callbacks.EarlyStopping(
        monitor='val_AUC', 
        #verbose=1,
        patience=1000,
        restore_best_weights=True)
    
    
    mcp_save = ModelCheckpoint('models/keras_classification.hdf5', save_best_only=True, monitor='val_AUC', mode='max')
    
    model = make_model()
    
    
    baseline_history = model.fit(
        X_train_s,
        y_train.astype('int'),
        verbose=1,
        batch_size=BATCH_SIZE,
        epochs=epochs,
        callbacks=[early_stopping, mcp_save],
        validation_data=(X_test_s, y_test.astype('float')))
    
    y_pred = model.predict(X_train_s)
    
    fpr, tpr, thresholds = roc_curve(y_train, y_pred)
    # calculate the g-mean for each threshold
    gmeans = np.sqrt(tpr * (1-fpr))
    # locate the index of the largest g-mean
    ix = np.argmax(gmeans)
    print('Best Threshold=%f, G-Mean=%.3f' % (thresholds[ix], gmeans[ix]))
    
    return model, thresholds[ix]
